///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  SubSpaceTracker.cpp                                                               //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:71b9596b-f8af-45b7-bbbb-7199f0ea0ce3>

#include  "modules.h"
#include  "Exception.h"
#include  "EFAPI.h"
#include  "{VAL label}.h"


{VAL label}::{VAL label}(void) :
{VAL params/input_rfifo_Label}({VAL params/input_fifo_ref}),
FRT_In({VAL params/input_rfifo_Label}),
PostFilt_Dimensions  { {VAL params/output_fifo_length}, {VAL params/num_channels},
                       {VAL params/FAPI_sample_len } + 1 },
PostFilt_WrapFlags  { true, false, false },
PostFilt_Extent  { PostFilt_Dimensions, PostFilt_WrapFlags },
{VAL params/output_fifo_label}(PostFilt_Extent),
FWT_Out({VAL params/output_fifo_label})
{
  this->State = State_Initialized;

  if( { VAL debug } )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


bool  {VAL label}::AnyActiveThreads(void) const
{
  bool  AnyActiveThreads = false;

  AnyActiveThreads = AnyActiveThreads || this->ControlThread.joinable();

  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    AnyActiveThreads = AnyActiveThreads || this->Thread[K].joinable();
  }

  return  AnyActiveThreads;
}


RTIF::ExitStateType  {VAL label}::Init_Allocate(void)
{
  this->State = State_Allocated;

  // Calculate 'NumProcThreads'
  if( this->NumProcThreads_FromParams <= this->NumChannels )
  {
    this->NumProcThreads =  this->NumProcThreads_FromParams;
  }
  else
  { // No need to spawn more threads than channels
    this->NumProcThreads = this->NumChannels;
  }

  // We need this value for allocating maximum array sizes during processing
  this->MaxChannelsPerProcThread = 0;

  // Determine which data channels are processed for each thread
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    this->BaseChannelNumberForThread( K,
                                      this->NumProcThreads,
                                      this->NumChannels,
                                      this->ChannelStart[K],
                                      this->NumChannelsPerThread[K] );

    if( this->NumChannelsPerThread[K] > this->MaxChannelsPerProcThread )
    {
      this->MaxChannelsPerProcThread = this->NumChannelsPerThread[K];
    }
  }

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // If there are active threads: kill them
    this->KillThreads();
    this->KillFlag = false;
    this->State = State_Allocated;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;

  if( { VAL debug } )
  { // Display Display_InitAlloc
    cout << Display_InitAlloc();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  {VAL label}::Init_Reset(void)
{
  this->State = State_Reset;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // There are active threads: kill them
    this->KillThreads();
    this->KillFlag = false;
    this->State = State_Reset;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;

  // Spawn the controller thread
  this->ControlThread = thread( this->ProcessBatches, ref(*this) );

  // Spawn the processing threads
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    this->ThreadCycleCounter[K] = ~0;
    this->Thread[K]             = thread( this->ProcessingThread, ref(*this), K );
  }

  { // Wait for all the threads to wake up and then wait on CycleCounter
    unique_lock<mutex>    LK(this->CV_Mutex);
    this->CV.wait( LK, [&]{ return (this->Alive == this->NumProcThreads) || this->KillFlag; } );
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


// Assuming all filters have the same number of coefficients
RTIF::ExitStateType  {VAL label}::ProcessBatches( {VAL label}& _Obj )
{
  uint32_t    WriteLen = 1024;
  uint32_t    ReadLen = 1024;

  do
  {
    // Wait for output space
    _Obj.{VAL params/output_fifo_label}.FIFO_Write_Start( WriteLen, _Obj.FWT_Out );

    // Wait for input data
    _Obj.{VAL params/input_rfifo_Label}.FIFO_Read_Fixed_Amount( ReadLen, _Obj.FRT_In );

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    { // Increment the cycle count to start processing
      unique_lock<mutex>   LK(_Obj.CV_Mutex);
      ++_Obj.CycleCounter;
      _Obj.Complete = 0;
      LK.unlock();
      _Obj.CV.notify_all();
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    /////////////////////////////////
    // Processing occurs here
    /////////////////////////////////

    { // Wait for children to finish
      unique_lock<mutex>   LK(_Obj.CV_Mutex);
      _Obj.CV.wait( LK, [&]{ return (_Obj.Complete == _Obj.NumProcThreads) || _Obj.KillFlag; } );
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Done writing into FIFO --- alert readers new space is available
    _Obj.{VAL params/output_fifo_label}.FIFO_Write_End(_Obj.FWT_Out);

    // Done writing into FIFO --- alert readers new space is available
    //   We are overlapping to 1 sample
    _Obj.{VAL params/input_rfifo_Label}.FIFO_Read_End( _Obj.FRT_In, 1 );

  }  while(!_Obj.KillFlag);

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::ProcessingThread( {VAL label}& _Obj, const uint32_t _ThreadNum )
{
  uint32_t  ChanOffset;
  const vector<uint32_t>  LocVect_In_Incr = { 1, 0 };
  const vector<uint32_t>  LocVect_Out_Incr = { 1, 0, 0 };
  vector<uint32_t>        LocVect_In = vector<uint32_t>(_Obj.{VAL params/input_rfifo_Label}.GetVectorLen());
  vector<uint32_t>        LocVect_Out = vector<uint32_t>(_Obj.{VAL params/output_fifo_label}.GetVectorLen());
  EFAPI_Real<int16_t,double,double,{VAL params/FAPI_sample_len},{VAL params/FAPI_rank}>
                                      EFapi[_Obj.NumChannelsPerThread[_ThreadNum]];

  // Grab local processing parameters
  _Obj.CV_Mutex.lock();
  const auto    LocChanBase   = _Obj.ChannelStart[_ThreadNum];
  const auto    LocNumChans   = _Obj.NumChannelsPerThread[_ThreadNum];
  _Obj.CV_Mutex.unlock();

  // Computational vectors[NumProcThreads][MaxChannelsPerProcThread][this->FftLen]
  AVX_Vector<AVX_Vector<{VAL params/input_type}>>
      BUF_IN( _Obj.MaxChannelsPerProcThread,
      AVX_Vector<{VAL params/input_type}>( {VAL params/FAPI_sample_len }+ 1, 0) );

  // Preset FAPI modules
  for( auto K=0U ; K < _Obj.NumChannelsPerThread[_ThreadNum] ; ++K )
  {
    EFapi[K].Init( {VAL params/FAPI_beta} );
  }

  { // Signal we're alive and ready to process data
    unique_lock<mutex>   LK(_Obj.CV_Mutex);
    ++_Obj.Alive;
    LK.unlock();
    _Obj.CV.notify_all();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return RTIF::Exit_Due_To_Kill;

  do
  {
    { // Wait for the cycle count to increment
      unique_lock<mutex>   LK(_Obj.CV_Mutex);
      _Obj.CV.wait( LK, [&]{ return (_Obj.CycleCounter != _Obj.ThreadCycleCounter[_ThreadNum]) || _Obj.KillFlag; } );
      ++_Obj.ThreadCycleCounter[_ThreadNum];
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    ///////////////////////
    // Over data channels
    ///////////////////////

    // Set-up and cache the Schemas
    const auto&  ReadSchema  = _Obj.FRT_In.GetReadSchema_View();
    const auto&  WriteSchema = _Obj.FWT_Out.GetWriteSchema_View();

    // Sync with FIFOs
    LocVect_In[0]  = ReadSchema.Start;
    LocVect_Out[0] = WriteSchema.Start;

    for( auto DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
    {
      ChanOffset = DataChan - LocChanBase;

      // Sync with FIFOs
      LocVect_In[1]  = DataChan;
      LocVect_Out[1] = DataChan;

      _Obj.{VAL params/input_rfifo_Label}.WriteFifoToVector<{VAL params/input_type}>
      (
        LocVect_In,
        LocVect_In_Incr,
        &BUF_IN[ChanOffset][0],
        ReadSchema.Length
      );

      // Process data vector with FAPI
      EFapi[ChanOffset].DataStep( BUF_IN[ChanOffset] );

      // Copy the data into the ouytput FIFO
      for( auto K=0U ; K < {VAL params/FAPI_rank} ; ++K )
      {
        LocVect_Out[2] = K;

        _Obj.{VAL params/output_fifo_label}.WriteVectorToFifo
        (
          &BUF_IN[ChanOffset][0],
          LocVect_Out,
          LocVect_Out_Incr,
          WriteSchema.Length
        );
      }
    }

    /////////////////////////////////
    // End of Processing
    /////////////////////////////////

    // Show that we are complete and ready to process another batch
    {
      unique_lock<mutex>   LK(_Obj.CV_Mutex);
      ++_Obj.Complete;
      LK.unlock();
      _Obj.CV.notify_all();
    }

  } while(!_Obj.KillFlag);

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::KillThreads(void)
{
  this->State = State_Kill;

  { // Set Kill flags in all FIFOs
    {VAL params/output_fifo_label}.SetKillFlag(true);
    this->{VAL params/input_rfifo_Label}.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(CV_Mutex);
    this->KillFlag = true;

    LK.unlock();

    CV.notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  // Now join up with the dying processing threads
  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    Thread[K].join();
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_Kill();
  }

  return  RTIF::Exit_Due_To_Kill;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  {VAL label}::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->CV_Mutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -NumProcThreads: " << this->NumProcThreads_FromParams << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;

  oS << "<< {VAL label}::Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  {VAL label}::Display_InitAlloc(void)
{
  lock_guard<mutex>   LK(this->CV_Mutex);
  ostringstream       oS;

  oS << ">> {VAL label}::InitAlloc[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Channels:" << endl;

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    oS << "    -Channel #: " << K << endl;
  }

  oS << "<< {VAL label}::InitAlloc[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->CV_Mutex);
  ostringstream       oS;

  oS << ">> {VAL label}::InitReset[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Channels:" << endl;

  for( auto K=0U ; K < this->NumChannels ; ++K )
  {
    oS << "    -Channel #: " << K << endl;
   }

  oS << "<< {VAL label}::InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->CV_Mutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Kill[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -Complete: " << this->Complete << endl;

  for( auto K=0U ; K < {VAL params/num_proc_threads} ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:"
       << endl;
   }

  oS << "<< {VAL label}::Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


