///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  DataPath.cpp                                                                      //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include  <iomanip>
#include  "DataPath.h"
#include  "DataNodeRef.h"
#include  "PARSER_Path.h"
#include  "ParseFrame.h"
#include  "Exception.h"

using namespace std;


// We only need one active parser
wej_parser::GrammarAbs*   DataPath::Parser  = nullptr;


DataPath::DataPath( const string& _Path ) :
TreeRoot(false)
{
  // Create the parser if we need it
  if( Parser == nullptr )
  {
    Parser  = path_bnf::Parse_Path();

    if(!Parser)
    {
      Exception   DE("Could not create 'DataPath' parser.",__FILE__,__LINE__);
      throw   DE;
    }
  }

  // Parse the path string
  ParseInPathString(_Path);
}


bool  DataPath::DowsNodeExist(void) const
{
  DataNodeRef   DNR( this->GetPathString(), false );
  return    DNR.DoesNodeExist();
}


void  DataPath::ParseInPathString( const string& _PathString )
{
  istringstream           IS( _PathString );
  wej_parser::ParseFrame  PF( IS, Out );

  if( !Parser->Parse(PF) )
  {
    Exception   DE(__FILE__,__LINE__);
    DE.Msg() << "Error parsing a 'DataPath': '" << _PathString << "'.";
    throw   DE;
  }

  // Set the root flag
  if( this->Out.Front().GetDataVect().operator[](0) == "D//" )
  {
    this->TreeRoot = true;
    this->Out.GetDataList().pop_front();
  }
  else if( this->Out.Front().GetDataVect().operator[](0) == "/" )
  {
    this->TreeRoot = false;
    this->Out.GetDataList().pop_front();
  }
  else
  {
    this->TreeRoot = false;
  }

  // If valid lock all data for our uses
  PF.AcceptAll();
}


bool    DataPath::IsASyntacticallyValidPath( const string& _Path )
{
  // Create the parser if we need it
  if( Parser == nullptr )
  {
    Parser  = path_bnf::Parse_Path();

    if(!Parser)
    {
      Exception   DE("Could not create 'DataPath' parser.",__FILE__,__LINE__);
      throw   DE;
    }
  }

  wej_parser::Outputs       OutLoc;
  istringstream             IS( _Path );
  wej_parser::ParseFrame    PF( IS, OutLoc );

  return  Parser->Parse(PF);
}


string  DataPath::GetPathString(void) const
{
  ostringstream   oS;

  if( this->TreeRoot )
  {
    oS << "//";
  }
  else
  {
    oS << "/";
  }

  for( auto I = this->Out.GetDataList().begin() ; I != this->Out.GetDataList().end() ; ++I )
  {
    if( I->GetDataVect()[0] != "UP" )
    {
      oS << I->GetDataVect()[1];
    }
    else
    {
      oS << "..";
    }

    auto J = I;
    ++J;
    if( J != this->Out.GetDataList().end() )
    {
      oS << '/';
    }
  }

  return  oS.str();
}


string   DataPath::DisplayAll( const unsigned int Indent ) const
{
  ostringstream   oS;

  for( auto K=0U ; K < Indent ; ++K )  oS << ' ';
  oS << "*DataRef: " << GetPathString() << endl;

  oS << Out.Display(Indent) << endl;

  return  oS.str();
}


DataPath   DataPath::operator+( const DataPath& _Data2 ) const
{
  DataPath     Ret;

  // Notes on operation:  If '_Data2' is a 'root' path (i.e. prefix //),
  // ------------------   the operation ignores the 'this' portion of the path.
  if( _Data2.IsTreeRoot() )
  {
    Ret  = _Data2;
  }
  else
  {
    // Local part
    for( const auto& I: this->Out.GetDataList() )
    {
      Ret.Out.GetDataList().push_back( wej_parser::OutputRec(I) );
    }

    // Data2 part
    for( const auto& I: _Data2.Out.GetDataList() )
    {
      Ret.Out.GetDataList().push_back( wej_parser::OutputRec(I) );
    }
  }

  Ret.Simplify();

  return      Ret;
}


void   DataPath::Simplify(void)
{
  unsigned int  Depth = 0;

  list<list<wej_parser::OutputRec>::iterator>   Excludes;

  for( auto I = this->Out.Begin() ; I != this->Out.End() ; ++I )
  {
    if( I->GetDataVect()[0] == "LAB" )
    {
      if( I->GetDataVect()[1] != ".." )
      {
        ++Depth;
      }
      else
      {
        if(Depth > 0)
        {
          Excludes.push_back(I);
          auto J = I;
          --J;
          Excludes.push_back(J);
          --Depth;
        }
      }
    }
  }

  // Remove the deleted records
  for( auto I=Excludes.begin() ; I!=Excludes.end() ; ++I )
  {
    this->Out.GetDataList().erase(*I);
  }
}


bool   DataPath::StripAndMatchPrefix( DataPath& _Prefix )
{
  list<list<wej_parser::OutputRec>::iterator>   Remove;

  auto  I = this->Out.Begin();
  auto  J = _Prefix.GetOutputList().begin();

  for(  ; (I != this->Out.End()) && (J != _Prefix.GetOutputList().end()) ; ++I, ++J )
  {
    if( !(I->GetDataVect()[1] == J->GetDataVect()[1]) )  return  false;

    Remove.push_back(I);
  }

  // Remove the matched items
  for( auto K=Remove.begin() ; K != Remove.end() ; ++K )
  {
    this->Out.GetDataList().erase(*K);
  }

  return  true;
}


