///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  SigProcWindows.h                                                                  //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <cstdint>
#include  <cmath>
#include  <numbers>
#include  <map>
#include  <vector>
#include  "AlignedMemAlloc.h"
#include  "Exception.h"

using namespace std;


///////////////////////////////////////////////////////////////////////////
//  SigProcWindows<typename DataType,uint32_t WindowLength>
///////////////////////////////////////////////////////////////////////////

template<typename DataType,uint32_t WindowLength>
class  SigProcWindows
{
private:

  typedef enum  eWindowTags
  {
    BLACKMAN_HARRIS_92
  }  WindowTagsType;

  map<string,WindowTagsType>  WindowNameMap
  {
    { "BlackmanHarris_92", BLACKMAN_HARRIS_92 }
  };

public:

  AVX_Vector<DataType>  GetSpectralWindow( const string& WindowNameParam )
  {
    // Look up the window
    const auto  LookUp = WindowNameMap.find(WindowNameParam);

    switch(LookUp->second)
    {
      case  BLACKMAN_HARRIS_92:
        return  BlackmanHarris_92();
      break;

      default:
      {
        Exception  E = Exception(__FILE__,__LINE__);
        E.Msg() << " Attempt to generate an undefined spectral window: '" << WindowNameParam << "'";
        throw  E;
      }
      break;
    }
  }

  AVX_Vector<DataType>  BlackmanHarris_92(void)
  {
    AVX_Vector<DataType>  Window(WindowLength);

    const double  A0 = .35875;
    const double  A1 = .48829;
    const double  A2 = .14128;
    const double  A3 = .01168;
    const double  TPIdivN = 2.*numbers::pi_v<DataType>/WindowLength;

    for( auto K=0U ; K < WindowLength ; ++K )
    {
      Window[K] = A0 - A1 * cos( TPIdivN * K )
                     + A2 * cos( TPIdivN * 2 * K )
                     - A3 * cos( TPIdivN * 3 * K );
    }

    return  Window;
  }
};


