///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  DataNodeRef.h                                                                     //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <string>
#include  "DataPath.h"
#include  "DataPool.h"
#include  "Exception.h"

using namespace std;


class  DataNodeRef
{
private:

  DataPath                  Path;
  DataPool::NodeRefType     Node;

  // Iterator variables
  list<wej_parser::XML_Node>::iterator  ChildIter;

public:

  /////////////////////////
  // Constructors
  /////////////////////////

  DataNodeRef( const DataPath& _DataPath, const bool _ThrowEx ) :
  Path(_DataPath),
  ChildIter(nullptr)
  {
    // Look up the path
    this->Node = wej_data_pool::DataTrees.FindPath(_DataPath);

    if( _ThrowEx && !this->Node.IsValid() )
    {
      throw   Exception
              (
                string("Data path '") +
                    _DataPath.GetPathString() +
                    string("' does not exist."),
                __FILE__, __LINE__
              );
    }
    else if( !this->Node.IsValid() )
    {
      cout << string(">> Data path '") +
                    _DataPath.GetPathString() +
                    string("' does not exist.") << endl;

    }

    if( this->Node.IsValid() && !IsRootLeaf() )
    {
      Begin();
    }
  }

  /////////////////////////
  // Field Accessors
  /////////////////////////

  inline const DataPath&   GetDataPath(void) const
  {
    return  Path;
  }

  inline const wej_parser::XML_Node&   GetXmlNode(void) const
  {
    if(Node.Node == nullptr)
    {
      throw   Exception
              (
                string("Node.Node is nullptr."),
                __FILE__, __LINE__
              );
    }

    return  *Node.Node;
  }

  inline wej_parser::XML_Node&   GetXmlNode(void)
  {
    if(Node.Node == nullptr)
    {
      throw   Exception
              (
                string("Node.Node is nullptr."),
                __FILE__, __LINE__
              );
    }

    return  *Node.Node;
  }

  /////////////////////////
  // Miscellaneous
  /////////////////////////

  bool    IsRootLeaf(void) const
  {
    return  GetXmlNode().IsLeaf();
  }

  bool    DoesNodeExist(void) const
  {
    return  this->Node.IsValid();
  }

  /////////////////////////
  // Child Iterator
  /////////////////////////

  inline void    Begin(void)
  {
    if(Node.Node == nullptr)
    {
      throw   Exception
              (
                string("Node.Node is nullptr."),
                __FILE__, __LINE__
              );
    }

    ChildIter  = Node.Node->ChildrenBegin();
  }

  inline void    Next(void)
  {
    ++ChildIter;
  }

  inline bool    MoreChildren(void) const
  {
    if(Node.Node == nullptr)
    {
      throw   Exception
              (
                string("Node.Node is nullptr."),
                __FILE__, __LINE__
              );
    }

    return    ChildIter != Node.Node->ChildrenEnd();
  }

  DataNodeRef   Curr(void) const;

  /////////////////////////
  // Display
  /////////////////////////

  string        Display( const unsigned int _Indent=0 ) const;
};


