///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  DataTree.cpp                                                                      //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include  "DataTree.h"
#include  "Exception.h"


DataTree::DataTree( const CD_ModesType  _Mode, const DataPath _DataPrefix ) :
Mode(_Mode),
FilePath(),
DataPrefix(_DataPrefix),
XML()
{ // Create a new empty one
  XML.GetHead().SetLabel("root");
}


DataTree::DataTree( const string& _FilePath, const CD_ModesType  _Mode, const DataPath _DataPrefix ) :
Mode(_Mode),
FilePath(_FilePath),
DataPrefix(_DataPrefix),
XML()
{
  if(!_DataPrefix.IsTreeRoot())
  {
    ostringstream oS;
    oS << "Data Prefix passed to DataTree Constructor is NOT a root node: '"
       << _DataPrefix.GetPathString() << "'.";
    throw   Exception(oS.str(),__FILE__,__LINE__);
  }

  if( (_Mode == eMODE_LOAD_FROM_FILE) || (_Mode == eMODE_LOAD_AND_SAVE_TO_FILE) )
  { // Read from an existing file
    bool Valid  = XML.ReadFromXML_File(_FilePath);

    if(!Valid)
    {
      ostringstream oS;
      oS << "Cannot open file '" << _DataPrefix.GetPathString()
         << "',  DataPath:" << _DataPrefix;
      throw   Exception(oS.str(),__FILE__,__LINE__);
    }
  }
  else
  { // Create a new empty one
    XML.GetHead().SetLabel("root");
  }
}


DataTree::~DataTree(void)
{
  if( (Mode == eMODE_SAVE_TO_FILE) || (Mode == eMODE_LOAD_AND_SAVE_TO_FILE) )
  {
    WriteFileToDisk();
  }
}


wej_parser::XML_Node*   DataTree::FindCreatePath( DataPath _Path )
{
  // First Simplify the path
  _Path.Simplify();

  // Check that the data storage prefix matches
  if( !_Path.StripAndMatchPrefix(DataPrefix) )   return  nullptr;

  // Simplify again
  _Path.Simplify();

  // Scan down the storage to find a slot to store the data in
  auto  N = &XML.GetHead();
  auto  I = _Path.GetOutputList().begin();

  // Search at a level
  for(  ; (I != _Path.GetOutputList().end()) && N ; ++I )
  {
    auto  N1 = N->LookUpNode(I->Data[1]);

    // See if we found it or not
    if( N1 )
    { // Existing
      N  = N1;
    }
    else
    { // Create a new node
      wej_parser::XML_Node    NewNode(N);

      NewNode.SetLabel(I->Data[1]);
      N->PushChild( NewNode );
      N = &N->ChildrenBack();
    }
  }

  return  N;
}


wej_parser::XML_Node*   DataTree::FindPath( DataPath _Path )
{
  // First Simplify the path
  _Path.Simplify();

  // Check that the data storage prefix matches
  if( !_Path.StripAndMatchPrefix(DataPrefix) )   return  nullptr;

  // Simplify again
  _Path.Simplify();

  // Scan down the storage to find a slot to store the data in
  auto  N = &XML.GetHead();
  auto  I = _Path.GetOutputList().begin();

  // Search at a level
  for(  ; (I != _Path.GetOutputList().end()) && N ; ++I )
  {
    auto  N1 = N->LookUpNode(I->Data[1]);

    // See if we found it or not
    if( N1 )
    { // Existing
      N  = N1;
    }
    else
    { // Cannot find path
      N = nullptr;
      break;
    }
  }

  return  N;
}


