///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  XML_Node.cpp                                                                      //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <sstream>
#include    "XML_Storage.h"
#include    "Exception.h"


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// string   XML_Node::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string  XML_Node::Display( const unsigned int Indent ) const
{
  ostringstream     oS, oS1;
  unsigned int      K;

  for( K=0 ; K < Indent ; ++K )   oS << ' ';
  oS1 << "/" << this->Label << "/";
  oS  << setw(20) << setiosflags(ios::left) << oS1.str() << endl;

  // Display the Attributes pairs
  for( auto I = Attributes.begin() ; I != Attributes.end() ; ++I )
  {
    for( K=0 ; K < Indent ; ++K )   oS << ' ';
    oS << "  Attr[" << I->Label << "]  '" << I->Value << "'" << endl;
  }

  // Display the Label:Value pairs
  for( auto J = Children.begin() ; J != Children.end() ; ++J )
  { // Display the child Nodes
    oS << J->Display( Indent+2 );
  }

  // Display the data
  if( !Data.empty() )
  {
    for( K=0 ; K < (Indent+2) ; ++K )   oS << ' ';
    oS << "'" << Data << "'" << endl;
  }

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   XML_Node::WriteXML
///////////////////////////////////////////////////////////////////////////////////////////////////

string  XML_Node::WriteXML( const unsigned int Indent, const bool Compact ) const
{
  ostringstream     oS;

  // Open node
  for( auto K=0U ; !Compact && (K < Indent) ; ++K )   oS << ' ';
  oS << "<" << this->GetLabel();

  // Display the attributes
  for( auto I=this->AttributesBegin() ; I != this->AttributesEnd() ; ++I )
  {
    oS << " " << I->Label << "=\"" << I->Value << "\"";
  }

  if( !Data.empty() )
  { // Data
    oS << ">";

    // Close and new line
    oS << XML_Node::ToGlyphs(this->Data);

    // Close the node
    oS << "</" << this->GetLabel() << ">";
    if(!Compact)  oS << endl;
  }
  else if( !this->Children.empty() )
  { // Children
    oS << ">";

    if(!Compact)  oS << endl;

    // Display the children
    for( auto J=this->ChildrenBegin() ; J != this->ChildrenEnd() ; ++J )
    {
      oS << J->WriteXML( Indent+2, Compact );
    }

    // Close the node
    for( auto K=0U ; !Compact && (K < Indent) ; ++K )   oS << ' ';
    oS << "</" << this->GetLabel() << ">";
    if(!Compact)  oS << endl;
  }
  else
  {
    // Empty Node
    oS << "/>";
    if(!Compact)  oS << endl;
  }

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   XML_Node::ToGlyphs
///////////////////////////////////////////////////////////////////////////////////////////////////

string    XML_Node::ToGlyphs( const string& In )
{
  string        R;
  unsigned int  K;

  // Over the characters
  for( K=0 ; K < In.size() ; ++K )
  {
    switch( In[K] )
    {
      case  '<':
        R   += "&lt;";
      break;

      case  '>':
        R   += "&gt;";
      break;

      case  '&':
        R  += "&amp;";
      break;

      case  '\'':
        R   += "&apos;";
      break;

      case  '\"':
        R   += "&quot;";
      break;

      default:
        R   += In[K];
      break;
    }
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Node*  XML_Node::LookUpNodeWithException
///////////////////////////////////////////////////////////////////////////////////////////////////

XML_Node*   XML_Node::LookUpNodeWithException( string XmlPath )
{
  XML_Node*                   N = this;
  size_t                      P;
  string                      SegLabel;
  XML_Node::ChildrenIterator  I;

  // Over the substrings
  do
  {
    // Isolate the next segment
    P  = XmlPath.find_first_of('.');

    // Isolate the segment Label
    SegLabel = XmlPath.substr( 0, P );

    // Look-Up the Local Label
    for( I=N->ChildrenBegin() ; I != N->ChildrenEnd() ; ++I )
    {
      if( I->GetLabel() == SegLabel )   break;
    }

    // See if we found the label
    if( I == N->ChildrenEnd() )
    {
      ostringstream   oS;
      oS << "Cannot find node '" << XmlPath << "'";
      throw  Exception(oS.str(),__FILE__,__LINE__);
    }

    // Set the pointer
    N  = &*I;

    // See if we are done
    if( P == string::npos )   break;

    // Get the next Section
    XmlPath   = XmlPath.substr( P+1 );

  } while( P != string::npos );

  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// const XML_Node*  XML_Node::LookUpNodeWithException
///////////////////////////////////////////////////////////////////////////////////////////////////

const XML_Node*   XML_Node::LookUpNodeWithException( string XmlPath ) const
{
  const XML_Node*                   N = this;
  size_t                            P;
  string                            SegLabel;
  XML_Node::ChildrenConstIterator   I;

  // Over the substrings
  do
  {
    // Isolate the next segment
    P   = XmlPath.find_first_of('.');

    // Isolate the segment Label
    SegLabel  = XmlPath.substr( 0, P );

    // Look-Up the Local Label
    for( I=N->ChildrenBegin() ; I != N->ChildrenEnd() ; ++I )
    {
      if( I->GetLabel() == SegLabel )   break;
    }

    // See if we found the label
    if( I == N->ChildrenEnd() )
    {
      ostringstream   oS;
      oS << "Cannot find node '" << XmlPath << "'";
      throw  Exception(oS.str(),__FILE__,__LINE__);
    }

    // Set the pointer
    N   = &*I;

    // See if we are done
    if( P == string::npos )   break;

    // Get the next Section
    XmlPath   = XmlPath.substr( P+1 );

  } while( P != string::npos );

  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Node*  XML_Node::LookUpNode
///////////////////////////////////////////////////////////////////////////////////////////////////

XML_Node*   XML_Node::LookUpNode( string XmlPath )
{
  XML_Node*                   N = this;
  size_t                      P;
  string                      SegLabel;
  XML_Node::ChildrenIterator  I;

  // Over the substrings
  do
  {
    // Isolate the next segment
    P   = XmlPath.find_first_of('.');

    // Isolate the segment Label
    SegLabel  = XmlPath.substr( 0, P );

    // Look-Up the Local Label
    for( I=N->ChildrenBegin() ; I != N->ChildrenEnd() ; ++I )
    {
      if( I->GetLabel() == SegLabel )  break;
    }

    // See if we found the label
    if( I == N->ChildrenEnd() )  return  nullptr;     // Cannot find node

    // Set the pointer
    N   = &*I;

    // See if we are done
    if( P == string::npos )  break;

    // Get the next Section
    XmlPath  = XmlPath.substr( P+1 );

  } while( P != string::npos );

  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// const XML_Node*  XML_Node::LookUpNode
///////////////////////////////////////////////////////////////////////////////////////////////////

const XML_Node*   XML_Node::LookUpNode( string XmlPath ) const
{
  const XML_Node*                   N = this;
  size_t                            P;
  string                            SegLabel;
  XML_Node::ChildrenConstIterator   I;

  // Over the substrings
  do
  {
    // Isolate the next segment
    P   = XmlPath.find_first_of('.');

    // Isolate the segment Label
    SegLabel  = XmlPath.substr( 0, P );

    // Look-Up the Local Label
    for( I=N->ChildrenBegin() ; I != N->ChildrenEnd() ; ++I )
    {
      if( I->GetLabel() == SegLabel )   break;
    }

    // See if we found the label
    if( I == N->ChildrenEnd() )   return nullptr;

    // Set the pointer
    N   = &*I;

    // See if we are done
    if( P == string::npos )   break;

    // Get the next Section
    XmlPath   = XmlPath.substr( P+1 );

  } while( P != string::npos );

  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// int    XML_Node::GetDataAsInt
///////////////////////////////////////////////////////////////////////////////////////////////////

int   XML_Node::GetDataAsInt(void) const
{
  int     N;

  // If data is nullptr assume it is 0
  if( Data.size() == 0 )  return 0;

  istringstream     iS(Data);
  iS >> N;
  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// unsigned int   XML_Node::GetDataAsUnsignedInt
///////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int      XML_Node::GetDataAsUnsignedInt(void) const
{
  unsigned int    N;

  // If data is nullptr assume it is 0
  if( Data.size() == 0 )  return 0;

  istringstream     iS(Data);
  iS >> N;
  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// long int     XML_Node::GetDataAsLongInt
///////////////////////////////////////////////////////////////////////////////////////////////////

long int      XML_Node::GetDataAsLongInt(void) const
{
  long int    N;

  // If data is nullptr assume it is 0
  if( Data.size() == 0 )  return 0;

  istringstream     iS(Data);
  iS >> N;
  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// unsigned long int    XML_Node::GetDataAsUnsignedLongInt
///////////////////////////////////////////////////////////////////////////////////////////////////

unsigned long int      XML_Node::GetDataAsUnsignedLongInt(void) const
{
  unsigned long int    N;

  // If data is nullptr assume it is 0
  if( Data.size() == 0 )  return 0;

  istringstream     iS(Data);
  iS >> N;
  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// double      XML_Node::GetDataAsDouble
///////////////////////////////////////////////////////////////////////////////////////////////////

double      XML_Node::GetDataAsDouble(void) const
{
  double    N;

  // If data is nullptr assume it is 0
  if( Data.size() == 0 )  return 0;

  istringstream     iS(Data);
  iS >> N;
  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// float      XML_Node::GetDataAsFloat
///////////////////////////////////////////////////////////////////////////////////////////////////

float      XML_Node::GetDataAsFloat(void) const
{
  float    N;

  // If data is nullptr assume it is 0
  if( Data.size() == 0 )  return 0;

  istringstream     iS(Data);
  iS >> N;
  return    N;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool      XML_Node::GetDataAsBool
///////////////////////////////////////////////////////////////////////////////////////////////////

bool      XML_Node::GetDataAsBool(void) const
{
  int    N;

  // If data is nullptr assume it is FALSE
  if( Data.size() == 0 )  return false;

  istringstream     iS(Data);
  iS >> N;
  return    N != 0;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string     XML_Node::GetXML_Path
///////////////////////////////////////////////////////////////////////////////////////////////////

string     XML_Node::GetXML_Path(void) const
{
  string      S;

  if( this->Father->Father )
  {
    S +=  this->Father->GetXML_Path();
  }

  S +=  '/';
  S +=  this->Label;

  return    S;
}
}   // End namespace 'wej_parser'


