///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  XML_Storage.cpp                                                                   //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include      <cassert>
#include      <sstream>
#include      <iostream>
#include      <iomanip>
#include      <fstream>
#include      "GrammarObj.h"
#include      "ParseFrame.h"
#include      "XML_Storage.h"
#include      "XML_Storage_Gram.h"


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// Binary file control bytes
///////////////////////////////////////////////////////////////////////////////////////////////////

const unsigned char   XML_Storage::SOF_Byte_0   = 0xA1;
const unsigned char   XML_Storage::SOF_Byte_1   = 0xF8;
const unsigned char   XML_Storage::SOF_Byte_2   = 0xED;
const unsigned char   XML_Storage::SOF_Byte_3   = 0x2B;
const unsigned char   XML_Storage::EOF_Byte_0   = 0xD6;
const unsigned char   XML_Storage::EOF_Byte_1   = 0xAF;
const unsigned char   XML_Storage::EOF_Byte_2   = 0x2D;
const unsigned char   XML_Storage::EOF_Byte_3   = 0x5C;
const unsigned char   XML_Storage::RandomSeed   = 0x79;


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::XML_Storage
///////////////////////////////////////////////////////////////////////////////////////////////////

XML_Storage::XML_Storage(void) : Valid(true), Head(nullptr)
{
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::ReadFromXML_File
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    XML_Storage::ReadFromXML_File( const string&  XML_FilePath )
{
  // Path to the XML source file
  ifstream      IN2( XML_FilePath.c_str(), ios::binary );

  // Make sure the file is open
  if( !IN2.is_open() )  return  false;

  // The output list
  Outputs       OutList;

  // Set up the Parse Frame
  ParseFrame    PF( IN2, OutList );

  // Create a C++ version of the XML parser
  GrammarAbs*   GA = wej_xml_parser::Parse_XML();

  // XML: XML_FilePath --> OutList
  bool    Ret   = GA->Parse(PF);

  // Make sure it got translated
  if( !Ret )
  {
    delete    GA;
    return  false;
  }

  // Parse the output and create the node list
  list<OutputRec>::const_iterator   I = OutList.Begin();

  // Recurse on the nodes
  bool    RetStatus   = ParseNode( this->Head, I, OutList, true );

  // Mark as valid
  Valid   = RetStatus;

  // Delete the grammar
  delete    GA;

  if(!Valid)  return  false;

  // At this point 'root' is one level lower.

  // Good return
  return  true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool    XML_Storage::ParseNode
///////////////////////////////////////////////////////////////////////////////////////////////////

bool          XML_Storage::ParseNode
              (
                XML_Node&                           NodeParam,
                list<OutputRec>::const_iterator&    I,
                const Outputs&                      OuputsParam,
                const bool                          AtHead
              )
{
  string        Label;
  bool          LastXmlDec = false;

  // Over the output tokens
  for(  ; I != OuputsParam.End() ; ++I )
  { // There must be a label
    assert( !I->GetDataVect().empty() );

    if( I->GetDataVect()[0] == "XML_DEC" )
    {
      LastXmlDec = true;
    }
    else if( I->GetDataVect()[0] == "NODE_LABEL" )
    {
      assert( I->GetDataVect().size() > 1 );
      Label   = I->GetDataVect()[1];
    }
    else if( I->GetDataVect()[0] == "NODE_STRING" )
    {
      assert( I->GetDataVect().size() > 1 );

      if(LastXmlDec)
      {
        this->PushXmlDecsLabelValuePair( Label, I->GetDataVect()[1] );
      }
      else
      {
        NodeParam.PushAttribute( Label, I->GetDataVect()[1] );
      }
    }
    else if( I->GetDataVect()[0] == "NODE_START" )
    {
      assert( I->GetDataVect().size() > 1 );
      LastXmlDec = false;

      // See if this is the head node 'AtHead=true'
      if( AtHead )
      { // At Head Node
        // Save the tag
        NodeParam.SetLabel( I->GetDataVect()[1] );

        // Parse the node
        ParseNode( NodeParam, ++I, OuputsParam, false );
      }
      else
      { // Normal Node --- Create a new child node
        NodeParam.PushChild( XML_Node(&NodeParam) );

        // Save the tag
        NodeParam.ChildrenBack().SetLabel( I->GetDataVect()[1] );

        // Parse the node
        ParseNode( NodeParam.ChildrenBack(), ++I, OuputsParam, false );
      }
    }
    else if( I->GetDataVect()[0] == "NODE_EMPTY" )
    {
      assert( I->GetDataVect().size() > 1 );
      LastXmlDec = false;

      // See if we are at the head
      if( AtHead )
      { // Save the tag
        NodeParam.SetLabel( I->GetDataVect()[1] );

        // Parse the node
        ParseNode( NodeParam, ++I, OuputsParam, false );
      }
      else
      { // Create a new node
        NodeParam.PushChild( XML_Node(&NodeParam) );

        // Save the tag
        NodeParam.ChildrenBack().SetLabel( I->GetDataVect()[1] );

        // Parse the node
        ParseNode( NodeParam.ChildrenBack(), ++I, OuputsParam, false );
      }
    }
    else if( I->GetDataVect()[0] == "NODE_EMPTY_DONE" )
    {
      return  true;
    }
    else if( I->GetDataVect()[0] == "NODE_DATA" )
    {
      assert( I->GetDataVect().size() > 1 );
      NodeParam.SetData( I->GetDataVect()[1] );
    }
    else if( I->GetDataVect()[0] == "NODE_END" )
    {
      assert( I->GetDataVect().size() > 1 );

      // Make sure the label matches the start label
      if( I->GetDataVect()[1] != NodeParam.GetLabel()  )  return  false;

      return  true;
    }
    else if( I->GetDataVect()[0] == "NODE_ERROR" )
    {
      return  false;
    }
    else
    {
      cout << ">> UNKNOWN: " <<  I->GetDataVect()[0] << endl;
    }
  }

  return    true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string     XML_Storage::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string     XML_Storage::Display(void) const
{
  ostringstream   oS;

  oS << "XML_Storage" << endl;
  oS << "{" << endl;

  // Display the Xml Declarations
  for( auto I=XmlDeclBegin() ; I != XmlDeclEnd() ; ++I )
  {
    oS << "  XML Dec: [']" << I->Label << "]  '" << I->Value << "'" << endl;
  }

  // Over the children
  for( auto J=Head.ChildrenBegin() ; J != Head.ChildrenEnd() ; ++J )
  {
    oS << J->Display(0);
  }

  // Display the nodes

  oS << "}" << endl;

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string    XML_Storage::GetXML_AsSourceString
///////////////////////////////////////////////////////////////////////////////////////////////////

string    XML_Storage::GetXML_AsSourceString( const bool Compact ) const
{
  ostringstream     oS;

  // Write the XML declaration line
  oS << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
  if(!Compact)  oS << endl;

  // Write the nodes
  oS << this->Head.WriteXML( 0, Compact );

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool      XML_Storage::WriteToXML_File
///////////////////////////////////////////////////////////////////////////////////////////////////

bool      XML_Storage::WriteToXML_File( const string& FilePath ) const
{
  ofstream    OUT( FilePath.c_str(), ios::binary | ios::trunc );

  // See if it is open
  if( !OUT.is_open() )  return  false;

  // Write the XML out
  OUT << GetXML_AsSourceString( XML_Storage::UseCompactXML_Format );

  // Good return
  return    true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool      XML_Storage::WriteToBinFile
///////////////////////////////////////////////////////////////////////////////////////////////////

bool      XML_Storage::WriteToBinFile( const string& FilePath ) const
{
  ofstream        Out( FilePath.c_str(), ios::binary | ios::trunc );
  F_RAND          Rand { XML_Storage::RandomSeed };
  unsigned char   UC;

  // See if it is open
  if( !Out.is_open() )  return  false;

  // Write a Start byte
  UC  = ApplyRand( XML_Storage::SOF_Byte_0, Rand );
  Out.write( (char*) &UC, 1 );

  UC  = ApplyRand( XML_Storage::SOF_Byte_1, Rand );
  Out.write( (char*) &UC, 1 );

  UC  = ApplyRand( XML_Storage::SOF_Byte_2, Rand );
  Out.write( (char*) &UC, 1 );

  UC  = ApplyRand( XML_Storage::SOF_Byte_3, Rand );
  Out.write( (char*) &UC, 1 );

  // Start writing the file
  if( !this->WriteBinLevel( this->GetHead(), Out, Rand ) ) return false;

  // Write a EOF byte
  UC  = ApplyRand( XML_Storage::EOF_Byte_0, Rand );
  Out.write( (char*) &UC, 1 );

  UC  = ApplyRand( XML_Storage::EOF_Byte_1, Rand );
  Out.write( (char*) &UC, 1 );

  UC  = ApplyRand( XML_Storage::EOF_Byte_2, Rand );
  Out.write( (char*) &UC, 1 );

  UC  = ApplyRand( XML_Storage::EOF_Byte_3, Rand );
  Out.write( (char*) &UC, 1 );

  // Good return
  return    true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool      XML_Storage::WriteBinLevel
///////////////////////////////////////////////////////////////////////////////////////////////////

bool      XML_Storage::WriteBinLevel
          (
            const XML_Node&     Node,
            ostream&            Out,
            F_RAND&             Rand
          ) const
{
  string    S;

  // Recover the Label and Lata
  unsigned char   LenLabel[2] = { (unsigned char) (Node.GetLabel().size() >> 8),
                                  (unsigned char)  Node.GetLabel().size()       };
  unsigned char   LenData[2]  = { (unsigned char) (Node.GetData().size() >> 8),
                                  (unsigned char)  Node.GetData().size()        };

  // Look for errors
  if( Node.GetLabel().size() > 0xFFFFU ) return false;
  if( Node.GetData().size()  > 0xFFFFU ) return false;

  // Save the length of the Label
  ApplyRand( LenLabel, 2, Rand );
  Out.write( (char*) LenLabel, 2 );

  // Save the Label
  S   = ApplyRand( Node.GetLabel(), Rand );
  Out.write( S.c_str(), S.size() );

  // Save the length of the Data
  ApplyRand( LenData, 2, Rand );
  Out.write( (char*) LenData, 2 );

  // Save the Data
  S   = ApplyRand( Node.GetData(), Rand );
  Out.write( S.c_str(), S.size() );

  // Save number of attributes
  unsigned char   NumAttrib[2] = { (unsigned char) (Node.GetNumAttributes() >> 8),
                                   (unsigned char)  Node.GetNumAttributes()     };
  ApplyRand( NumAttrib, 2, Rand );
  Out.write( (char*) NumAttrib, 2 );

  // Over the node attributes
  for( auto I=Node.AttributesBegin() ; I != Node.AttributesEnd() ; ++I )
  {
    // Look for errors
    if( I->Label.size() > 0xFFFFU ) return false;
    if( I->Value.size() > 0xFFFFU ) return false;

    // Recover the Label and Data
    unsigned char   LenLabel[2] = { (unsigned char) (I->Label.size() >> 8),
                                    (unsigned char)  I->Label.size()      };
    unsigned char   LenData[2]  = { (unsigned char) (I->Value.size() >> 8),
                                    (unsigned char)  I->Value.size()      };

    // Save the length of the Label
    ApplyRand( LenLabel, 2, Rand );
    Out.write( (char*) LenLabel, 2 );

    // Save the Label
    S   = ApplyRand( I->Label, Rand );
    Out.write( S.c_str(), S.size() );

    // Save the length of the Data
    ApplyRand( LenData, 2, Rand );
    Out.write( (char*) LenData, 2 );

    // Save the Data
    S   = ApplyRand( I->Value, Rand );
    Out.write( S.c_str(), S.size() );
  }

  // Save number of children
  unsigned char   NumChild[2] = { (unsigned char) (Node.GetNumChildren() >> 8),
                                  (unsigned char)  Node.GetNumChildren()     };

  ApplyRand( NumChild, 2, Rand );
  Out.write( (char*) NumChild, 2);

  // Over the children of this node
  for( auto I=Node.ChildrenBegin() ; I != Node.ChildrenEnd() ; ++I )
  {
    WriteBinLevel( *I, Out, Rand );
  }

  return  true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::ReadFromBinFile
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    XML_Storage::ReadFromBinFile( const string&  ScenFilePath )
{
  ifstream          In( ScenFilePath.c_str(), ios::binary );
  F_RAND            Rand { XML_Storage::RandomSeed };
  unsigned char     Buf;

  // See if it is open
  if( !In.is_open() )  return  false;

  // Read a SOF byte
  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::SOF_Byte_0 )  return false;

  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::SOF_Byte_1 )  return false;

  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::SOF_Byte_2 )  return false;

  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::SOF_Byte_3 )  return false;

  // Clear the local tree so we can rebuild
  this->GetHead().Clear();

  // Push a child to start the process
  this->GetHead().PushChild( XML_Node(&this->GetHead()) );

  // Start writing the file
  if( !this->ReadBinLevel( this->GetHead().ChildrenBack(), In, Rand ) ) return false;

  // Read a EOF byte
  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::EOF_Byte_0 )  return false;

  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::EOF_Byte_1 )  return false;

  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::EOF_Byte_2 )  return false;

  In.read( (char*) &Buf, 1 );
  if( In.gcount() != 1 )  return false;
  if( ApplyRand( Buf, Rand ) != XML_Storage::EOF_Byte_3 )  return false;

  // Good return
  return    true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::ReadBinLevel
///////////////////////////////////////////////////////////////////////////////////////////////////

bool      XML_Storage::ReadBinLevel
          (
            XML_Node&       Node,
            istream&        In,
            F_RAND&         Rand
          )
{
  unsigned char   Buf[2];

  // Recover the Label length
  In.read( (char*) &Buf, 2 );
  if( In.gcount() != 2 ) return false;

  unsigned int    LabelLength = ( ApplyRand( Buf[0], Rand ) << 8 ) |
                                  ApplyRand( Buf[1], Rand          );

  unsigned char*  LabelBuf = new unsigned char[LabelLength];

  // Recover the Label Data
  In.read( (char*) LabelBuf, LabelLength );
  if( (unsigned) In.gcount() != LabelLength ) return false;

  // Recover the Data length
  In.read( (char*) &Buf, 2 );
  if( In.gcount() != 2  ) return false;
  ApplyRand( LabelBuf, LabelLength, Rand );

  unsigned int    DataLength  = ( ApplyRand( Buf[0], Rand ) << 8 ) |
                                  ApplyRand( Buf[1], Rand          );

  unsigned char*  DataBuf = new unsigned char[DataLength];

  // Recover the Data
  In.read( (char*) DataBuf, DataLength );
  if( (unsigned) In.gcount() != DataLength ) return false;
  ApplyRand( DataBuf,  DataLength,  Rand );

  // Save the data in the node
  Node.SetLabel( string( (char*) LabelBuf, LabelLength ) );
  Node.SetData ( string( (char*) DataBuf,  DataLength  ) );

  // Deallocate the memory
  delete[]  LabelBuf;
  delete[]  DataBuf;

  // Get the number of attributes
  In.read( (char*) Buf, 2 );
  if( (unsigned) In.gcount() != 2 ) return false;

  unsigned int    NumAttributes = ( (ApplyRand( Buf[0], Rand ) << 8) |
                                     ApplyRand( Buf[1], Rand )       );

  // Process any attributes
  for( auto K=0U ; K < NumAttributes ; ++K )
  {
    // Recover the Label length
    In.read( (char*) &Buf, 2 );
    if( In.gcount() != 2 ) return false;
    unsigned int  LabelLength = ( (ApplyRand( Buf[0], Rand ) << 8) |
                                   ApplyRand( Buf[1], Rand )       );

    unsigned char*  LabelBuf = new unsigned char[LabelLength];

    // Recover the Label Data
    In.read( (char*) LabelBuf, LabelLength );
    if( (unsigned) In.gcount() != LabelLength ) return false;
    ApplyRand( LabelBuf, LabelLength, Rand );

    // Recover the Data length
    In.read( (char*) &Buf, 2 );
    if( In.gcount() != 2  ) return false;
    unsigned int    DataLength  = ( (ApplyRand( Buf[0], Rand ) << 8) |
                                     ApplyRand( Buf[1], Rand )       );

    unsigned char*  DataBuf = new unsigned char[DataLength];

    // Recover the Label Data
    In.read( (char*) DataBuf, DataLength );
    if( (unsigned) In.gcount() != DataLength ) return false;
    ApplyRand( DataBuf,  DataLength,  Rand );

    // Create the attribute
    Node.PushAttribute( string( (char*) LabelBuf, LabelLength ),
                        string( (char*) DataBuf,  DataLength  ) );

    // Deallocate the memory
    delete[]  LabelBuf;
    delete[]  DataBuf;
  }

  // Get the number of children
  In.read( (char*) Buf, 2 );
  if( (unsigned) In.gcount() != 2 ) return false;

  unsigned int    NumChildren = ( ApplyRand( Buf[0], Rand ) << 8) |
                                  ApplyRand( Buf[1],Rand )        ;

  // Process any children
  for( auto K=0U ; K < NumChildren ; ++K )
  {
    // Push a new node
    Node.PushChild( XML_Node(&Node) );

    // Process the new node
    if( !ReadBinLevel( Node.ChildrenBack(), In, Rand ) ) return false;
  }

  return  true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::ApplyRand
///////////////////////////////////////////////////////////////////////////////////////////////////

void    XML_Storage::ApplyRand( unsigned char* Buf, const unsigned int Length, F_RAND& Rand )
{
  for( auto K=0U ; K < Length ; ++K )
  {
    Buf[K] ^=  Rand();
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::ApplyRand
///////////////////////////////////////////////////////////////////////////////////////////////////

unsigned char   XML_Storage::ApplyRand( unsigned char Buf, F_RAND& Rand )
{
  return    Buf ^ Rand();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// XML_Storage::ApplyRand
///////////////////////////////////////////////////////////////////////////////////////////////////

string    XML_Storage::ApplyRand( const string& Buf, F_RAND& Rand )
{
  string    OBuf( Buf );

  for( auto K=0U ; K < Buf.size() ; ++K )
  {
    OBuf[K] ^=  Rand();
  }

  return    OBuf;
}
}   // End namespace 'wej_parser'


